-------------------------------------------------------------------------------
-- SPDX-FileCopyrightText: 2022 CERN (home.cern)
--
-- SPDX-License-Identifier: CERN-OHL-W-2.0+
-------------------------------------------------------------------------------
-- Title      : mt_profip_translator
-- Project    : ProFIP
-------------------------------------------------------------------------------
-- File       : mt_profip_translator.vhd
-- Author     : Konstantinos Blantos
-- Company    : CERN (BE-CEM-EDL)
-- Created    : 06-12-2022
-- Last update:
-- Platform   : FPGA-generic
-- Standard   : VHDL'08
-------------------------------------------------------------------------------
-- Description: Responsible of received the MOSI data from ERTEC through SPI
--              and transmitting MISO data from Mockturtle. For communication
--              with MockTurtle, Remote Message Queues are used.
-------------------------------------------------------------------------------

--=============================================================================
--                            Libraries & Packages                           --
--=============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.mock_turtle_pkg.all;
use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.mt_mqueue_pkg.all;

--=============================================================================
--               Entity declaration for mt_profip_translator                 --
--=============================================================================

entity mt_profip_translator is
generic (
  g_data_width     : natural;
  g_cpol           : natural;
  g_cpha           : natural);
port (
  clk_i            : in  std_logic;  -- FPGA clock
  rst_n_i          : in  std_logic;  -- FPGA reset
--  ertec_dbg_o      : out std_logic_vector(6 downto 0);
--  clk_ref_i        : in std_logic;

  -- RX (MSTRFIP TRANSLATOR -> RMQ)
  rmq_src_i        : in  t_mt_stream_source_in;  -- (ready, pkt_ready)
  rmq_src_o        : out t_mt_stream_source_out; -- (data, hdr, valid, last, error)
  rmq_src_config_i : in  t_mt_stream_config_out; -- (adr, dat, we)
  rmq_src_config_o : out t_mt_stream_config_in;  -- (dat)

  -- TX (RMQ -> MSTRFIP_TRANSLATOR)
  rmq_snk_i        : in  t_mt_stream_sink_in;    -- (data, hdr, valid, last, error)
  rmq_snk_o        : out t_mt_stream_sink_out;   -- (ready, pkt_ready)
  rmq_snk_config_i : in  t_mt_stream_config_out; -- (adr, dat, we)
  rmq_snk_config_o : out t_mt_stream_config_in;  -- (dat)

  -- SPI interface
  spi_clk_i        : in  std_logic;  -- SPI clock
  spi_cs_n_i       : in  std_logic;  -- SPI chip select, active LOW
  spi_mosi_i       : in  std_logic;  -- SPI serial data, master out slave in
  spi_miso_o       : out std_logic); -- SPI serial data, master in slave out
end entity mt_profip_translator;

--==============================================================================
--                           Architecture declaration                         --
--==============================================================================

architecture rtl of mt_profip_translator is

  -- FSM types
  type t_wr_state is (IDLE, BASIC_INFO, HEADER_1, HEADER_2, WAIT_HEADER,
                   DATA_1, DATA_2, WAIT_DATA, CHECKSUM);

  type t_rd_state is (IDLE, RD_HEADER, PRE_LOAD, LOAD_HWORD, LOAD_LWORD, DATA);

  signal s_wr_state : t_wr_state;
  signal s_rd_state : t_rd_state;

  -- Internal signals
  signal spi_mosi_data      : std_logic_vector(g_data_width-1 downto 0) := (others=>'0');
  signal spi_miso_data      : std_logic_vector(g_data_width-1 downto 0) := (others=>'0');
  signal mosi_data          : std_logic_vector(g_data_width-1 downto 0) := (others=>'0');
  signal wr_control         : std_logic_vector(7 downto 0); -- stores the CONTROL byte from MOSI
  signal wr_rmq_id          : std_logic_vector(7 downto 0); -- stores the RMQ_ID  byte from MOSI
  signal wr_data_len        : std_logic_vector(7 downto 0); -- stores the LENGTH  byte from MOSI
  signal wr_data_cnt        : unsigned(7 downto 0); -- counts the number of data bytes from MOSI
  signal wr_header_cnt      : unsigned(1 downto 0); -- counts the number of header bytes from MOSI
  signal wr_header          : std_logic_vector(15 downto 0);
  signal wr_data            : std_logic_vector(15 downto 0);
  signal spi_data_valid_i   : std_logic;
  signal spi_data_valid_o   : std_logic;
  signal valid              : std_logic;
  signal fail               : std_logic;
  signal rd_en              : std_logic;
  --signal rd_en_cnt          : unsigned(1 downto 0) := (others=>'0');
  signal spi_busy           : std_logic;
  signal spi_ready          : std_logic;
  signal miso_data_reg      : std_logic_vector(31 downto 0);
  signal rd_header_data     : std_logic_vector(g_data_width+3 downto 0) := (others=>'0'); -- reg to store temporary the output valid data of MISO FIFO
  -- MOSI async FIFO signals
  signal mosi_fifo_data     : std_logic_vector(g_data_width-1 downto 0) := (others=>'0');
  signal mosi_wr_en         : std_logic;
  signal mosi_rd_en         : std_logic;
  signal mosi_fifo_wr_empty : std_logic;
  signal mosi_fifo_wr_full  : std_logic;
  signal mosi_fifo_rd_empty : std_logic;
  signal mosi_fifo_rd_full  : std_logic;
  -- MISO async FIFO signals
  signal miso_fifo_data_in  : std_logic_vector(g_data_width+3 downto 0) := (others=>'0');
  signal miso_fifo_data_out : std_logic_vector(g_data_width+3 downto 0) := (others=>'0');
  signal miso_wr_en         : std_logic;
  signal miso_rd_en         : std_logic;
  signal miso_fifo_wr_empty : std_logic;
  signal miso_fifo_wr_full  : std_logic;
  signal miso_fifo_rd_empty : std_logic;
  signal miso_fifo_rd_full  : std_logic;
  signal miso_last          : std_logic;

  signal s_data_valid       : std_logic;

begin

  ----------------------------------------------------------------------------
  -- Included components in MasterFIP Translator
  ----------------------------------------------------------------------------
  cmp_spi_slave : entity work.spi_slave
  generic map (
    g_cpol       => g_cpol,
    g_cpha       => g_cpha,
    g_data_width => g_data_width)
  port map (
    clk_i        => clk_i, --clk_ref_i,
    rst_n_i      => rst_n_i,
    spi_clk_i    => spi_clk_i,    -- spi_clk_domain
    spi_cs_n_i   => spi_cs_n_i,   -- spi_clk_domain
    spi_mosi_i   => spi_mosi_i,   -- spi_clk_domain
    spi_miso_o   => spi_miso_o,   -- spi_clk_domain
    data_i       => spi_miso_data,
    data_valid_i => spi_data_valid_i,
    data_o       => spi_mosi_data,
    data_valid_o => spi_data_valid_o,
    ready_o      => spi_ready,
    busy_o       => spi_busy);

  -- Asynchronous FIFO used for fixing the CDC between
  -- SPI Slave and Translator, because there are both
  -- working under different clock domains.
  cmp_mosi_async_fifo : entity work.inferred_async_fifo
  generic map (
    g_data_width             => g_data_width,
    g_size                   => 32,
    g_show_ahead             => TRUE,
    g_with_rd_empty          => TRUE,
    g_with_rd_full           => FALSE,
    g_with_rd_almost_empty   => FALSE,
    g_with_rd_almost_full    => FALSE,
    g_with_rd_count          => FALSE,
    g_with_wr_empty          => FALSE,
    g_with_wr_full           => TRUE,
    g_with_wr_almost_empty   => FALSE,
    g_with_wr_almost_full    => FALSE,
    g_with_wr_count          => FALSE,
    g_almost_empty_threshold => 0,
    g_almost_full_threshold  => 1023)
  port map (
    rst_n_i           => rst_n_i,
    clk_wr_i          => clk_i, --clk_ref_i,
    d_i               => spi_mosi_data,
    we_i              => mosi_wr_en,
    wr_empty_o        => mosi_fifo_wr_empty,
    wr_full_o         => mosi_fifo_wr_full,
    wr_almost_empty_o => open,
    wr_almost_full_o  => open,
    wr_count_o        => open,
    clk_rd_i          => clk_i,
    q_o               => mosi_fifo_data,
    rd_i              => mosi_rd_en,
    rd_empty_o        => mosi_fifo_rd_empty,
    rd_full_o         => mosi_fifo_rd_full,
    rd_almost_empty_o => open,
    rd_almost_full_o  => open,
    rd_count_o        => open);

  -- Asynchronous FIFO used for fixing the CDC between
  -- Translator and SPI Slave, because there are both
  -- working under different clock domains. Data width
  -- of this FIFO will be different because now data
  -- contains the control signals (hdr,valid,error,last)
  cmp_miso_async_fifo : entity work.inferred_async_fifo
  generic map (
    g_data_width             => g_data_width+4,
    g_size                   => 32,
    g_show_ahead             => TRUE,
    g_with_rd_empty          => TRUE,
    g_with_rd_full           => FALSE,
    g_with_rd_almost_empty   => FALSE,
    g_with_rd_almost_full    => FALSE,
    g_with_rd_count          => FALSE,
    g_with_wr_empty          => FALSE,
    g_with_wr_full           => TRUE,
    g_with_wr_almost_empty   => FALSE,
    g_with_wr_almost_full    => FALSE,
    g_with_wr_count          => FALSE,
    g_almost_empty_threshold => 0,
    g_almost_full_threshold  => 1023)
  port map (
    rst_n_i           => rst_n_i,
    clk_wr_i          => clk_i,
    d_i               => miso_fifo_data_in, -- contains control signals too
    we_i              => miso_wr_en,
    wr_empty_o        => miso_fifo_wr_empty,
    wr_full_o         => miso_fifo_wr_full,
    wr_almost_empty_o => open,
    wr_almost_full_o  => open,
    wr_count_o        => open,
    clk_rd_i          => clk_i, --clk_ref_i,
    q_o               => miso_fifo_data_out,
    rd_i              => miso_rd_en,
    rd_empty_o        => miso_fifo_rd_empty,
    rd_full_o         => miso_fifo_rd_full,
    rd_almost_empty_o => open,
    rd_almost_full_o  => open,
    rd_count_o        => open);

  -- t_mt_stream_config_in and out will not be used
  rmq_src_config_o <= c_MT_DUMMY_EP_CONFIG_IN;
  rmq_snk_config_o <= c_MT_DUMMY_EP_CONFIG_IN;

  -- Final and synchronized MOSI data that will go
  -- to RMQ is the data taken from SPI Slave when
  -- the read enable of MOSI FIFO is HIGH
--  mosi_data <= mosi_fifo_data when mosi_rd_en = '1';

  process(clk_i,rst_n_i)
  begin
    if rst_n_i = '0' then
      mosi_data <= (others=>'0');
    elsif rising_edge(clk_i) then
      if mosi_rd_en = '1' then
        mosi_data <= mosi_fifo_data;
      else
        mosi_data <= mosi_data;
      end if;
    end if;
  end process;

  -- The format of the data is the following:
  -- 35   - hdr
  -- 34   - valid
  -- 33   - last
  -- 32   - error
  -- 31:0 - data
  miso_fifo_data_in <= (rmq_snk_i.hdr & rmq_snk_i.valid & rmq_snk_i.last & rmq_snk_i.error & rmq_snk_i.data); --it is in 62.5MHz

  ----------------------------------------------------------------------------
  -- RX:  MASTERFIP TRANSLATOR --> MT (RMQ)
  ----------------------------------------------------------------------------

  -- FSM for writing from ERTEC to RMQ. Exiting the IDLE state when source
  -- input is ready and there is valid data coming from MOSI FIFO. In the BASIC
  -- INFO state the received data consist of 3 bytes (1 control byte, 1 RMQ_ID
  -- byte and 1 byte containg the Number of Bytes that the data contains). The
  -- In the HEADER state the header bytes assigned to src_o.data (3x32-bits).
  -- 4th byte is unused. Depending on the data len, received in the BASIC INFO,
  -- there is a counter which counts the number of packets of 32-bit data that
  -- received and is to be sent to RMQ. Once it is stops transmitting, last
  -- is asserted. CHECKSUM state, is the state that FSM will go for 1 clock cycle,
  -- and error will be asserted. Then, either it is an error or not, FSM will IDLE.
  -- NOTE: the incoming mosi data has 32-bit width, same as rmq data bus, but
  -- rmq discard half of them everytime, so that's why valid data should be send
  -- as a packet of 16-bits every time.
  p_wr_fsm : process(clk_i,rst_n_i,spi_cs_n_i)
  begin
    if (rst_n_i = '0') then
      s_wr_state    <= IDLE;
      wr_control    <= (others=>'0');
      wr_rmq_id     <= (others=>'0');
      wr_data_len   <= (others=>'0');
      wr_header_cnt <= (others=>'1');
      wr_data_cnt   <= (others=>'0');
      fail          <= '0';
    elsif rising_edge(clk_i) then
      case s_wr_state is

        when IDLE =>
          if (rmq_src_i.ready = '1' and mosi_rd_en = '1') then
            s_wr_state    <= BASIC_INFO;
            wr_control    <= mosi_fifo_data(31 downto 24);
            wr_rmq_id     <= mosi_fifo_data(23 downto 16);
            wr_data_len   <= mosi_fifo_data(15 downto 8);
            wr_header_cnt <= (others=>'1');
          else
            s_wr_state    <= IDLE;
            wr_control    <= (others=>'0');
            wr_rmq_id     <= (others=>'0');
            wr_data_len   <= (others=>'0');
            wr_header_cnt <= (others=>'1');
            wr_data_cnt   <= (others=>'0');
            fail          <= '0';
          end if;

        when BASIC_INFO =>
          if (rmq_src_i.ready = '1' and mosi_rd_en = '1') then
            wr_data_cnt <= unsigned(wr_data_len); -- assign the width of payload in bytes
            s_wr_state  <= HEADER_1;
          elsif (spi_cs_n_i = '1') then
            s_wr_state <= CHECKSUM;
            fail       <= '1';
          else
            s_wr_state <= BASIC_INFO;
            fail       <= '0';
          end if;

        when HEADER_1 =>
          if (rmq_src_i.ready = '1') then
            if wr_header_cnt /= 0 then
              s_wr_state    <= HEADER_2;
              wr_header_cnt <= wr_header_cnt - 1;
            else
              s_wr_state    <= IDLE; --otherwise we will be stuck for ever here
            end if;
          elsif (spi_cs_n_i = '1') then
            fail       <= '1';
            s_wr_state <= CHECKSUM;
          else
            s_wr_state <= HEADER_1;
            fail       <= '0';
          end if;

        when HEADER_2 =>
          if (rmq_src_i.ready = '1') then
            s_wr_state <= WAIT_HEADER;
          elsif (spi_cs_n_i = '1') then
            fail       <= '1';
            s_wr_state <= CHECKSUM;
          else
            fail       <= '0';
            s_wr_state <= HEADER_2;
          end if;

        when WAIT_HEADER =>
          if (rmq_src_i.ready = '1' and mosi_rd_en = '1') then
            if wr_header_cnt = 0 then
              s_wr_state <= DATA_1;
            else
              s_wr_state <= HEADER_1;
            end if;
          elsif (spi_cs_n_i = '1') then
            fail <= '1';
            s_wr_state <= CHECKSUM;
          else
            fail       <= '0';
            s_wr_state <= WAIT_HEADER;
          end if;

        when DATA_1 =>
          if (rmq_src_i.ready = '1') then
            if wr_data_cnt /=0 then
              s_wr_state <= DATA_2;
              wr_data_cnt <= wr_data_cnt - 1;
            else
              s_wr_state <= IDLE; --otherwise we will be stuck for ever here
            end if;
          elsif (spi_cs_n_i = '1') then
            fail       <= '1';
            s_wr_state <= CHECKSUM;
          else
            s_wr_state <= DATA_1;
            fail       <= '0';
          end if;

        when DATA_2 =>
          if (rmq_src_i.ready = '1') then
            s_wr_state <= WAIT_DATA;
          elsif (spi_cs_n_i = '1') then
            fail       <= '1';
            s_wr_state <= CHECKSUM;
          else
            fail       <= '0';
            s_wr_state <= DATA_2;
          end if;

        when WAIT_DATA =>
          if (rmq_src_i.ready = '1' and mosi_rd_en = '1') then
            if wr_data_cnt = 0 then
              s_wr_state <= CHECKSUM;
            else
              s_wr_state <= DATA_1;
            end if;

          elsif (wr_data_cnt = 0) then
            s_wr_state <= CHECKSUM;
          else
            fail <= '0';
            s_wr_state <= WAIT_DATA;
          end if;

        when CHECKSUM =>
          fail <= '0';
          if spi_cs_n_i = '1' then
            s_wr_state <= IDLE;
          else
            s_wr_state <= CHECKSUM;
          end if;

        when others =>
          fail       <= '0';
          s_wr_state <= IDLE;

      end case;
    end if;
  end process p_wr_fsm;


  -- RMQ protocol, requires that we should devide the header and payload data into 2 parts
  wr_header <= mosi_data(31 downto 16) when s_wr_state = HEADER_1 else
               mosi_data(15 downto 0)  when s_wr_state = HEADER_2 else (others=>'0');

  wr_data   <= mosi_data(31 downto 16) when s_wr_state = DATA_1 else
               mosi_data(15 downto 0)  when s_wr_state = DATA_2 else (others=>'0');

  rmq_src_o.data <= (x"0000" & wr_header) when (s_wr_state = HEADER_1 or s_wr_state = HEADER_2) else
                    (x"0000" & wr_data) when (s_wr_state = DATA_1 or s_wr_state = DATA_2) else (others=>'0');

  rmq_src_o.hdr  <= '1' when (s_wr_state = HEADER_1 or s_wr_state = HEADER_2) else '0';
  valid <= '1' when ((s_wr_state = HEADER_1 or s_wr_state = HEADER_2) or (s_wr_state = DATA_1 or s_wr_state = DATA_2)) else '0';
  rmq_src_o.valid <= valid or fail; -- in case to catch the error, we need valid HIGH (!)
  rmq_src_o.last <= '1' when (wr_data_cnt = 0 and valid = '1') else '0';

  -- Assert error signal when less header or payload data has been sent by ERTEC (and CS is HIGH)
  rmq_src_o.error <= fail;

  -- Write enable of the MOSI FIFO
  mosi_wr_en <= '1' when (spi_data_valid_o = '1' and mosi_fifo_wr_full ='0') else '0'; -- we are using the output valid from SPI Slave (it is in the clk domain of wr)

  -- Read enable of the MOSI FIFO
  mosi_rd_en <= '1' when (mosi_fifo_rd_empty = '0' and rmq_src_i.ready = '1') else '0';

  ----------------------------------------------------------------------------
  -- TX: MT (RMQ) --> MASTERFIP TRANSLATOR
  ----------------------------------------------------------------------------

  -- FSM for reading data from RMQ. The basic idea is that we read the data from
  -- the MISO FIFO that it is already stored. If it is a header data
    p_rd_fsm : process(clk_i, rst_n_i) --clk_ref_i, rst_n_i) --spi_clk domain
    begin
        if rst_n_i = '0' then
            s_rd_state    <= IDLE;
            rd_en         <= '0';
            miso_data_reg <= (others=>'0');
            s_data_valid  <= '0';
            rd_header_data <= (others => '0');
            miso_last <= '0';

        elsif rising_edge(clk_i) then --clk_ref_i) then
            case s_rd_state is
                ----------------------------------------------------------------
                --! STATE IDLE
                when IDLE =>
                    miso_last <= '0';
                    if spi_ready = '1' and miso_fifo_rd_empty = '0' then
                        rd_en      <= '1';
                        s_data_valid <= '1';
                        miso_data_reg <= miso_fifo_data_out(31 downto 0);
                        s_rd_state <= RD_HEADER;
                    else
                        miso_data_reg <= miso_data_reg;
                        s_data_valid <= '0';
                        rd_en <= '0';
                        s_rd_state <= IDLE;
                    end if;

                ----------------------------------------------------------------
                --! STATE RD_HEADER
                when RD_HEADER =>
                    rd_en      <= '0';
                    s_data_valid <= '0';
                    s_rd_state <= PRE_LOAD;

                ----------------------------------------------------------------
                --! STATE PRE_LOAD
                when PRE_LOAD =>
                    s_data_valid <= '0';
                    if miso_fifo_rd_empty = '0' then
                        rd_en      <= '1';
                        s_rd_state <= LOAD_HWORD;
                    else
                        rd_en      <= '0';
                        s_rd_state <= PRE_LOAD;
                    end if;

                ----------------------------------------------------------------
                --! STATE LOAD_HWORD
                when LOAD_HWORD =>
                    s_data_valid <= '0';
                    if miso_fifo_rd_empty = '0' then
                        rd_en <= '1';
                        if miso_fifo_data_out(34) = '1' and miso_fifo_data_out(32) = '0' then
                            miso_data_reg(31 downto 16) <= miso_fifo_data_out(15 downto 0);
                            s_rd_state <= LOAD_LWORD;
                        else
                            s_rd_state <= IDLE;
                        end if;
                    else
                        rd_en <= '0';
                        s_rd_state <= LOAD_HWORD;
                    end if;

                ----------------------------------------------------------------
                --! STATE LOAD_LWORD
                when LOAD_LWORD =>
                    if miso_fifo_rd_empty = '0' then
                        rd_en <= '1';
                        if miso_fifo_data_out(34) = '1' and miso_fifo_data_out(32) = '0' then
                            miso_data_reg(15 downto 0) <= miso_fifo_data_out(15 downto 0);
                            if miso_fifo_data_out(33) = '1' then
                                miso_last <= '1';
                            end if;
                            s_rd_state <= DATA;
                        else
                            s_rd_state <= IDLE;
                        end if;
                    else
                        rd_en <= '0';
                        s_rd_state <= LOAD_LWORD;
                    end if;

                ----------------------------------------------------------------
                --! STATE DATA
                when DATA =>
                    rd_en <= '0';
                    if spi_ready = '1' then
                        s_data_valid <= '1';
                        if miso_last = '1' then
                            s_rd_state <= IDLE;
                        else
                            s_rd_state <= LOAD_HWORD;
                        end if;
                    else
                        s_rd_state <= DATA;
                    end if;

                ----------------------------------------------------------------
                --! STATE DEFAULT
                when others =>
                    s_rd_state <= IDLE;
                    rd_en <= '0';

                ----------------------------------------------------------------
            end case;
        end if;
    end process;

  -- Ready and Packet Ready
  rmq_snk_o.ready     <= '1'; -- when last = '1' we are not ready
  rmq_snk_o.pkt_ready <= '1';

  -- Write enable of the MISO FIFO
  miso_wr_en <= '1' when (rmq_snk_i.valid = '1' and rmq_snk_i.error = '0' and miso_fifo_wr_full = '0') else '0';

  -- Read enable of the MISO FIFO
  miso_rd_en <= rd_en;
  spi_miso_data <= miso_data_reg;
  spi_data_valid_i <= s_data_valid;

end architecture;

end architecture rtl;

